///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

/*
 * \file BinaryStream.h
 * \author tsharpe
 * \date Aug 12, 2009
 *
 * \brief Facility for serializing data structures to a file.
 * This is an incredibly naive version that doesn't handle any of the niceties
 * such as object cycles, version compatibility, etc.
 */
#ifndef FEUDAL_BINARYSTREAM_H_
#define FEUDAL_BINARYSTREAM_H_

#include "feudal/BinaryStreamTraits.h"
#include "system/Assert.h"
#include <fstream>
#include <ostream>
#include <istream>
#include <string.h>
#include <iterator>
#include <utility>
#include <string>
#include <vector>

TRIVIALLY_SERIALIZABLE(bool);
TRIVIALLY_SERIALIZABLE(char);
TRIVIALLY_SERIALIZABLE(signed char);
TRIVIALLY_SERIALIZABLE(unsigned char);
TRIVIALLY_SERIALIZABLE(short);
TRIVIALLY_SERIALIZABLE(unsigned short);
TRIVIALLY_SERIALIZABLE(int);
TRIVIALLY_SERIALIZABLE(unsigned int);
TRIVIALLY_SERIALIZABLE(long);
TRIVIALLY_SERIALIZABLE(unsigned long);
TRIVIALLY_SERIALIZABLE(float);
TRIVIALLY_SERIALIZABLE(double);

class MagicToken
{
public:
    MagicToken()
    { memcpy(mToken,"BINWRITER",sizeof(mToken)); }

    // compiler-supplied destructor and copying are OK

    bool isValid()
    { return !memcmp(mToken,"BINWRITER",sizeof(mToken)); }

private:
    char mToken[8];
};

TRIVIALLY_SERIALIZABLE(MagicToken);

/// Writer of binary streams.
class BinaryWriter
{
public:
    BinaryWriter( char const* fileName )
    : mFileName(fileName),
      mOFS(fileName,std::ios_base::out|std::ios_base::binary|std::ios_base::trunc),
      mOS(mOFS)
    { if ( !mOFS ) fail(); write(MagicToken()); }

    BinaryWriter( std::ostream& os )
    : mFileName("(unknown stream)"), mOS(os)
    { write(MagicToken()); }

    /// Constructor that allows you to replace the "magic token" header
    /// with an alternate of your choosing (or none, if hdrLen = 0).
    BinaryWriter( std::ostream& os, void const* header, size_t hdrLen )
    : mOS(os)
    { writeBuf(reinterpret_cast<char const*>(header),hdrLen); }

    /// Constructor that allows you to replace the "magic token" header
    /// with an alternate of your choosing (or none, if hdrLen = 0).
    BinaryWriter( char const* fileName, void const* header, size_t hdrLen )
    : mOFS(fileName,std::ios_base::out|std::ios_base::binary|std::ios_base::trunc), mOS(mOFS)
    { writeBuf(reinterpret_cast<char const*>(header),hdrLen); }

    ~BinaryWriter()
    { if ( mOFS.is_open() ) close(); }

    /// Write something.
    template <class T>
    BinaryWriter& write( T const& val )
    { writeVal(val,Serializability<T>()); return *this; }

    /// Write an array of somethings.
    template <class T>
    BinaryWriter& write( T const* begin, T const* end )
    { writeArray(begin,end,Serializability<T>()); return *this; }

    /// Write a sequence from an iterator pair.
    template <class Itr>
    BinaryWriter& writeItr( Itr begin, Itr const& end )
    { while ( begin != end ) { write(*begin); ++begin; } return *this; }

    void close();

    /// Write a single object to a file in a single, simple function call.
    template <class T>
    static void writeFile( char const* fileName, T const& obj )
    { BinaryWriter(fileName).write(obj).close(); }

protected:

    template <class SIZ_T>
    void patchSize( SIZ_T val )
    { mOFS.flush().seekp(sizeof(MagicToken)); writeBuf(&val,sizeof(val)); }

private:
    BinaryWriter( BinaryWriter const& ); // unimplemented -- no copying
    BinaryWriter& operator=( BinaryWriter const& ); // unimplemented -- no copying

    template <class T>
    void writeVal( T const& val, TriviallySerializable )
    { writeBuf(&val,sizeof(T)); }

    template <class T>
    void writeVal( T const& val, SelfSerializable )
    { val.writeBinary(*this); }

    template <class T>
    void writeVal( T const& val, ExternallySerializable )
    { writeBinary(*this,val); }

    template <class T>
    void writeArray( T const* begin, T const* end, TriviallySerializable )
    { writeBuf(begin,sizeof(T)*(end-begin)); }

    template <class T>
    void writeArray( T const* begin, T const* end, SelfSerializable )
    { while ( begin != end ) write(*begin++); }

    template <class T>
    void writeArray( T const* begin, T const* end, ExternallySerializable )
    { while ( begin != end ) write(*begin++); }

    void writeBuf( void const* data, size_t len )
    { if ( !mOS.write(static_cast<char const*>(data),len) ) fail(); }

    void fail();

    std::string mFileName;
    std::ofstream mOFS;
    std::ostream& mOS;
};

/// Reader of binary streams.
class BinaryReader
{
public:
    BinaryReader( char const* fileName )
    : mFileName(fileName),
      mIFS(fileName,std::ios_base::in|std::ios_base::binary),
      mIS(mIFS)
    { testToken(); }

    BinaryReader( std::istream& is, bool checkToken = true )
    : mFileName("(unknown stream)"), mIS(is)
    { if ( checkToken ) testToken(); }

    ~BinaryReader()
    { if ( mIFS.is_open() ) mIFS.close(); }

    /// Read something.
    template <class T>
    T& read( T* pT )
    { readVal(pT,Serializability<T>()); return *pT; }

    /// Read an array.
    template <class T>
    void read( T* begin, T* end )
    { readArray(begin,end,Serializability<T>()); }

    /// Read a sequence delimited by an iterator pair.
    template <class Itr>
    void readItr( Itr begin, Itr const& end )
    { while ( begin != end ) { read(&*begin); ++begin; } }

    /// Read a single object from a binary file in one step.
    template <class T>
    static T& readFile( char const* fileName, T* obj )
    { return BinaryReader(fileName).read(obj); }

    template <class T>
    static size_t externalSizeof( T* arg )
    { return externalSizeof(arg,Serializability<T>()); }

private:
    BinaryReader( BinaryReader const& ); // unimplemented -- no copying
    BinaryReader& operator=( BinaryReader const& ); // unimplemented -- no copying

    template <class T>
    void readVal( T* pT, TriviallySerializable )
    { doRead(pT,sizeof(T)); }

    /// Read something that handles its own serialization.
    template <class T>
    void readVal( T* pVal, SelfSerializable )
    { pVal->readBinary(*this); }

    template <class T>
    void readVal( T* pT, ExternallySerializable )
    { readBinary(pT,*this); }

    template <class T>
    void readArray( T* begin, T* end, TriviallySerializable )
    { doRead(begin,sizeof(T)*(end-begin)); }

    template <class T>
    void readArray( T* begin, T* end, SelfSerializable )
    { while ( begin != end ) read(begin++); }

    template <class T>
    void readArray( T* begin, T* end, ExternallySerializable )
    { while ( begin != end ) read(begin++); }

    void doRead( void* pVal, size_t len )
    { if ( !mIS.read(static_cast<char*>(pVal),len) ) fail(); }

    template <class T>
    static size_t externalSizeof( T*, TriviallySerializable )
    { return sizeof(T); }

    template <class T>
    static size_t externalSizeof( T*, SelfSerializable )
    { return T::externalSizeof(); }

    template <class T>
    static size_t externalSizeof( T* arg, ExternallySerializable )
    { return serializedSizeof(arg); }

    void testToken();
    void fail();

    std::string mFileName;
    std::ifstream mIFS;
    std::istream& mIS;
};

/// This is a very fragile class.  You parameterize it on some vector-like type,
/// and it allows you to read the elements of that vector from a binary stream
/// one by one without reconstituting the whole vector.
/// However:  It assumes that the vector was written to the stream as a
/// V::size_type followed by the V::value_type elements (as, indeed, all vectors
/// do, as of this writing).  If this isn't true, the whole thing will
/// likely explode in your face.
template <class V>
class BinaryIteratingReader
{
public:
    typedef typename V::size_type size_type;
    typedef typename V::value_type value_type;

    BinaryIteratingReader( char const* fileName )
    : mRdr(fileName)
    { mRdr.read(&mCount); }

    BinaryIteratingReader( std::istream& is )
    : mRdr(is)
    { mRdr.read(&mCount); }

    // compiler-supplied destructor is OK

    /// number of elements left to read
    size_type remaining() const { return mCount; }

    /// get the next element
    value_type next()
    { AssertGt(mCount,0u);
      --mCount;
      value_type result;
      mRdr.read(&result);
      return result; }

    class iterator
    : public std::iterator<std::input_iterator_tag,value_type>
    {
    public:
        iterator( BinaryIteratingReader* pIRdr = 0 )
        : mpIRdr(pIRdr)
        { operator++(); }

        // compiler-supplied destructor and copying are OK

        value_type const& operator*() const { Assert(mOK); return mVal; }
        value_type const* operator->() const { Assert(mOK); return &mVal; }

        iterator& operator++()
        { if ( (mOK = mpIRdr && mpIRdr->remaining()) ) mVal = mpIRdr->next();
          return *this; }

        iterator operator++( int )
        { iterator tmp(*this); operator++(); return tmp; }

        friend bool operator==( iterator const& itr1, iterator const& itr2 )
        { return itr1.mOK == itr2.mOK; }

        friend bool operator!=( iterator const& itr1, iterator const& itr2 )
        { return itr1.mOK != itr2.mOK; }

    private:
        BinaryIteratingReader* mpIRdr;
        value_type mVal;
        bool mOK; // indicates that mVal has a valid value
    };

    iterator begin() { return iterator(this); }
    iterator end() { return iterator(0); }

private:
    BinaryIteratingReader( BinaryIteratingReader const& ); // unimplemented -- no copying
    BinaryIteratingReader& operator=( BinaryIteratingReader const& ); // unimplemented -- no copying

    BinaryReader mRdr;
    typename V::size_type mCount;
};

/// Same, fragile idea as above.  Allow a vector-like type to be written
/// incrementally.  Closing patches the vector size, which is initially
/// recorded as 0.
template <class V>
class BinaryIterativeWriter : private BinaryWriter
{
public:
    BinaryIterativeWriter( char const* fileName )
    : BinaryWriter(fileName), mCount(0)
    { BinaryWriter::write(mCount); }

    BinaryIterativeWriter& write( typename V::value_type const& val )
    { BinaryWriter::write(val); mCount += 1; return *this; }

    void close()
    { BinaryWriter::patchSize(mCount); BinaryWriter::close(); }

private:
    typename V::size_type mCount;
};

/*
 * specializations for some commonly-used STL classes
 */
template <class T1, class T2>
struct Serializability<std::pair<T1,T2> > : public ExternallySerializable {};

/// Write a std::pair.
template <typename T1, typename T2>
inline void writeBinary( BinaryWriter& writer, std::pair<T1,T2> const& val )
{ writer.write(val.first); writer.write(val.second); }

/// Read a std::pair.
template <typename T1, typename T2>
inline void readBinary( std::pair<T1,T2>* pVal, BinaryReader& reader )
{ reader.read(&pVal->first); reader.read(&pVal->second); }

template <typename T1, typename T2>
inline size_t serializedSizeof( std::pair<T1,T2>* )
{ size_t sz1 = BinaryReader::externalSizeof(static_cast<T1*>(0));
  size_t sz2 = BinaryReader::externalSizeof(static_cast<T2*>(0));
  return sz1 && sz2 ? sz1+sz2 : 0UL; }

template <>
struct Serializability<std::string> : public ExternallySerializable {};

/// Write a std::string.
inline void writeBinary( BinaryWriter& writer, std::string const& str )
{ std::string::size_type nnn = str.size(); writer.write(nnn);
  char const* buf = &str[0]; writer.write(buf,buf+nnn); }

/// Read a std::string.
inline void readBinary( std::string* pStr, BinaryReader& reader )
{ std::string::size_type sz; reader.read(&sz); pStr->resize(sz);
  char* buf = &(*pStr)[0]; reader.read(buf,buf+sz); }

inline size_t serializedSizeof( std::string* )
{ return 0UL; }

template <class T, class A>
struct Serializability<std::vector<T,A> > : public ExternallySerializable {};

/// Write a std::vector
template <class T>
inline void writeBinary( BinaryWriter& writer, std::vector<T> const& vec )
{ typename std::vector<T>::size_type nnn = vec.size(); writer.write(nnn);
  T const* buf = &vec[0]; writer.write(buf,buf+nnn); }

/// Read a std::vector
template <class T>
inline void readBinary( std::vector<T>* pVec, BinaryReader& reader )
{ typename std::vector<T>::size_type sz; reader.read(&sz); pVec->resize(sz);
  T* buf = &(*pVec)[0]; reader.read(buf,buf+sz); }

template <class T>
inline size_t serializedSizeof( std::vector<T>* )
{ return 0UL; }

#endif /* FEUDAL_BINARYSTREAM_H_ */
