///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

/*
 * \file FeudalControlBlock.cc
 * \author tsharpe
 * \date Jun 23, 2009
 *
 * \brief
 */
#ifndef FORCE_DEBUG
#define NDEBUG
#endif
#include "feudal/FeudalControlBlock.h"
#include "system/Exit.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <iostream>

using std::cout;
using std::endl;

FeudalControlBlock::FeudalControlBlock( char const* fileName, bool validate, unsigned long* pFileLen )
{
    int fd = open(fileName,O_RDONLY);
    if ( fd == -1 )
    {
        cout << "Can't open Feudal File " << fileName << endl;
        CRD::exit(1);
    }
    ctor_guts(fd,validate,pFileLen);
    close(fd);
}

FeudalControlBlock::FeudalControlBlock( int fd, bool validate, unsigned long* pFileLen )
{
    ctor_guts(fd,validate,pFileLen);
}

bool FeudalControlBlock::isValid( unsigned long fileLen, bool verbose ) const
{
    if ( getNFiles() != 1 )
    {
        if ( verbose )
        {
            if ( getNFiles() )
                cout << "Feudal file is in " << getNFiles() <<
                        " files, but we require it to be in a single file." << endl;
            else
                cout << "Feudal file claims to exist in 0 files.  It's probably "<<
                        "not a feudal file, or the process of creating it was interrupted." << endl;
        }
        return false;
    }

    unsigned long offsetBytes = mFixedOffset - mVarOffset;
    if ( offsetBytes % sizeof(unsigned long) )
    {
        if ( verbose )
            cout << "Feudal file's offset info does not contain an integral number of offsets." << endl;
        return false;
    }

    //unsigned int nnn = offsetBytes/sizeof(unsigned long) - 1UL;
    if ( ((offsetBytes/sizeof(unsigned long) - 1UL) & 0xffffffffUL) != mN )
    {
        if ( verbose )
            cout << "Feudal file doesn't have the right number of offsets for the number of elements." << endl;
        return false;
    }

    if ( !isCompressed() )
    {
        unsigned long fixedBytes = fileLen - mFixedOffset;
        unsigned long nElements = getNElements();
        if ( !nElements )
        {
            if ( fixedBytes )
            {
                if ( verbose )
                    cout << "Empty feudal file has some fixed data.  Must be garbage at the file's end." << endl;
                return false;
            }
        }
        else
        {
            if ( fixedBytes % nElements )
            {
                if ( verbose )
                    cout << "Feudal file's fixed info is not an integral number of bytes per element." << endl;
                return false;
            }

            if ( mSizeofFixed && ((fixedBytes/nElements)&0xff) != mSizeofFixed )
            {
                if ( verbose )
                    cout << "Feudal file doesn't have the right amount of fixed data for the number of elements." << endl;
                return false;
            }
        }
    }

    return true;
}

bool FeudalControlBlock::isGoodFeudalFile( char const* fileName, bool verbose )
{
    unsigned long fileLen;
    FeudalControlBlock fcb(fileName,false,&fileLen);
    return fcb.isValid(fileLen,verbose);
}

void FeudalControlBlock::ctor_guts( int fd, bool validate, unsigned long* pFileLen )
{
    struct stat buf;
    if ( fstat(fd,&buf) )
    {
        if ( validate )
        {
            cout << "Can't fstat the feudal file: " << strerror(errno) << endl;
            CRD::exit(1);
        }
        buf.st_size = 0;
    }

    if ( pFileLen )
        *pFileLen = buf.st_size;

    if ( read(fd,this,sizeof(*this)) != sizeof(*this) )
    {
        if ( validate )
        {
            cout << "Can't read feudal file: " << strerror(errno) << endl;
            CRD::exit(1);
        }
        memset(this,0,sizeof(*this));
    }

    if ( validate && !isValid(buf.st_size,true) )
        CRD::exit(1);
}
