///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

/**
 * \file FeudalFileReader.cc
 * \author tsharpe
 * \date Wednesday, December 03, 2008
 *
 * \brief FeudalFileReader implementation
 */
#ifndef FORCE_DEBUG
#define NDEBUG
#endif
#include "feudal/FeudalFileReader.h"
#include "system/Assert.h"
#include "system/System.h"
#include <cstring>
#include <cerrno>
#include <unistd.h>
#include <sys/mman.h>

FeudalFileReader::FeudalFileReader( char const* filename )
: mFilename(filename), mReader(mIFS,false), mFCB(0,0,0,0,0,0)
{
    init();
}

FeudalFileReader::FeudalFileReader( FeudalFileReader const& that )
: mFilename(that.mFilename), mReader(mIFS,false), mFCB(0,0,0,0,0,0)
{
    init();
}

FeudalFileReader::~FeudalFileReader()
{
    uninit();
}

FeudalFileReader& FeudalFileReader::operator=( FeudalFileReader const& that )
{
    uninit();
    mFilename = that.mFilename;
    init();
    return *this;
}

size_t FeudalFileReader::getPreallocation( size_t maxSize, size_t extSize,
                                           size_t start, size_t end ) const
{
    AssertLe(end,getNElements());
    AssertLe(start,end);

    size_t result = 0;
    while ( start < end )
    {
        size_t nnn = getDataLen(start++)/extSize;
        if ( nnn <= maxSize )
            result += nnn;
    }
    return result;
}

void FeudalFileReader::init()
{
    ForceAssertEq(sizeof(size_t),8u);

    int fd = open(mFilename.c_str(),O_RDONLY);
    if ( fd == -1 )
        FatalErr("Unable to open " << mFilename << ": " << strerror(errno));

    size_t fileLen;
    new (&mFCB) FeudalControlBlock(fd,true,&fileLen);

    size_t varTabOffset = mFCB.getVarTabOffset();
    size_t mappedOffset = varTabOffset - varTabOffset%getpagesize();
    mMappedLen = fileLen - mappedOffset;
    mMappedBit = static_cast<char*>(mmap(0,mMappedLen,PROT_READ,MAP_SHARED,fd,mappedOffset));
    if ( mMappedBit == MAP_FAILED )
        FatalErr("Unable to mmap " << mFilename << ": " << strerror(errno));

    close(fd);

    mOffsetsTable = mMappedBit + (varTabOffset - mappedOffset);
    mFixedData = mMappedBit + (mFCB.getFixedOffset() - mappedOffset);

    mIFS.open(mFilename.c_str());
    mIFSPos = mNEle = mFCB.getNElements();
}

void FeudalFileReader::uninit()
{
    mIFS.close();
    munmap(mMappedBit,mMappedLen);
}
