///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

/*
 * \file FeudalFileWriter.cc
 * \author tsharpe
 * \date Mar 13, 2009
 *
 * \brief Utility for writing Feudal Files incrementally.
 */
#ifndef FORCE_DEBUG
#define NDEBUG
#endif
#include "feudal/FeudalFileWriter.h"
#include "feudal/FeudalControlBlock.h"
#include "system/Exit.h"
#include <string>
#include <string.h>
#include <errno.h>
#include <iostream>
using std::cout;
using std::endl;
using std::ios_base;
using std::vector;

FeudalControlBlock FeudalFileWriter::gInvalidFCB(0,0,0,0,0,0);

FeudalFileWriter::FeudalFileWriter( char const* filename,
                                    FeudalFileWriter::size_type vecSize,
                                    FeudalFileWriter::size_type eltSize,
                                    FeudalFileWriter::size_type fixedLenDataLen,
                                    unsigned long estimatedNElements )
: mFilename(filename),
  mOS(filename,ios_base::out|ios_base::trunc|ios_base::binary),
  mBinaryWriter(mOS,&gInvalidFCB,sizeof(FeudalControlBlock)),
  mNextOffset(mOS.tellp()),
  mVecSize(vecSize),
  mEltSize(eltSize),
  mFixedLenDataLen(fixedLenDataLen),
  mWasCheckpointed(false)
{
    mOffsets.reserve(estimatedNElements);
    mFixedLenData.reserve(fixedLenDataLen*estimatedNElements);
}

FeudalFileWriter::~FeudalFileWriter()
{
    // nothing to do. this is not inlined or defaulted because we don't want
    // the compiler-generated version to be inline:  there are too many object
    // destructions.
}

void FeudalFileWriter::addElement( void const* fixedLenData,
                                   void const* variableLenData,
                                   unsigned long variableLenDataNElts )
{
    if ( !mOS.is_open() )
    {
        cout << "You can't add more to the feudal file " << mFilename <<
                " after you've closed it." << endl;
        CRD::exit(1);
    }

    if ( mWasCheckpointed )
    {
        mOS.seekp(0UL);
        mBinaryWriter.write(gInvalidFCB);
        mOS.seekp(mNextOffset);
        mWasCheckpointed = false;
    }

    char const* vStart = reinterpret_cast<char const*>(variableLenData);
    mBinaryWriter.write(vStart,vStart + variableLenDataNElts*mEltSize);

    addElement( fixedLenData );
}

void FeudalFileWriter::addElement( void const* fixedLenData )
{
    mOffsets.push_back(mNextOffset);
    mNextOffset = mOS.tellp();

    if ( mFixedLenDataLen )
    {
        char const* fStart = reinterpret_cast<char const*>(fixedLenData);
        mFixedLenData.insert(mFixedLenData.end(),fStart,fStart+mFixedLenDataLen);
    }
}

void FeudalFileWriter::checkPoint()
{
    if ( !mOS.is_open() )
    {
        cout << "You can't checkpoint a FeudalFileWriter after you've closed it." << endl;
        CRD::exit(1);
    }
    finish();
    mOS.seekp(mNextOffset);
    mWasCheckpointed = true;
}

void FeudalFileWriter::close()
{
    if ( !mOS.is_open() )
        return; // EARLY RETURN!

    finish();
    mOS.close();
}

void FeudalFileWriter::finish()
{
    mBinaryWriter.write(&*mOffsets.begin(),&*mOffsets.end());
    mBinaryWriter.write(mNextOffset);
    mBinaryWriter.write(&*mFixedLenData.begin(),&*mFixedLenData.end());

    if ( !mOS.flush() )
    {
        cout << "Can't sync file " << mFilename << endl;
        CRD::exit(1);
    }

    mOS.seekp(0UL);
    FeudalControlBlock fcb(mOffsets.size(),
                           mNextOffset-sizeof(FeudalControlBlock),
                           mFixedLenDataLen, mVecSize, mEltSize);
    mBinaryWriter.write(fcb);
    if ( !mOS.flush() )
    {
        cout << "Can't sync file " << mFilename << endl;
        CRD::exit(1);
    }
}
