///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

/*
 * \file Iterator.h
 * \author tsharpe
 * \date Jun 9, 2009
 *
 * \brief Random-access iterators for vector-like classes.
 */
#ifndef FEUDAL_ITERATOR_H_
#define FEUDAL_ITERATOR_H_

#include <iterator>

/// An abstract iterator position.
template <class S>
class IteratorPosition
{
public:
    explicit IteratorPosition( S pos ) : mPos(pos) {}

    // compiler-supplied copying and destructor are OK

    S pos() const { return mPos; }

protected:
    S mPos;
};

/*
 * Iterator position comparisons.
 */

template <class S>
bool operator==( IteratorPosition<S> const& i1, IteratorPosition<S> const& i2 )
{ return i1.pos() == i2.pos(); }

template <class S>
bool operator!=( IteratorPosition<S> const& i1, IteratorPosition<S> const& i2 )
{ return i1.pos() != i2.pos(); }

template <class S>
bool operator<( IteratorPosition<S> const& i1, IteratorPosition<S> const& i2 )
{ return i1.pos() < i2.pos(); }

template <class S>
bool operator<=( IteratorPosition<S> const& i1, IteratorPosition<S> const& i2 )
{ return i1.pos() <= i2.pos(); }

template <class S>
bool operator>( IteratorPosition<S> const& i1, IteratorPosition<S> const& i2 )
{ return i1.pos() > i2.pos(); }

template <class S>
bool operator>=( IteratorPosition<S> const& i1, IteratorPosition<S> const& i2 )
{ return i1.pos() >= i2.pos(); }


/// A piece of a random-access iterator for vector-like classes.
/// This piece implements position movement.
template <class X,class S,class D>
class IteratorBase : public IteratorPosition<S>
{
public:
    IteratorBase() : IteratorPosition<S>(0) {}
    explicit IteratorBase( S pos ) : IteratorPosition<S>(pos) {}

    // compiler-supplied copying and destructor are OK

    X& operator++() { fwd(1); return thisX(); }
    X operator++(int) { X tmp(thisX()); fwd(1); return tmp; }
    X& operator+=( D diff ) { fwd(diff); return thisX(); }
    X operator+( D diff ) const { X tmp(thisX()); tmp.fwd(diff); return tmp; }
    X& operator--() { bwd(1); return thisX(); }
    X operator--(int) { X tmp(thisX()); bwd(1); return tmp; }
    X& operator-=( D diff ) { bwd(diff); return thisX(); }
    X operator-( D diff ) const { X tmp(thisX()); tmp.bwd(diff); return tmp; }

    D operator-( IteratorPosition<S> const& that )
    { return this->mPos - that.pos(); }

    void fwd( D diff ) { this->mPos += diff; }
    void bwd( D diff ) { this->mPos -= diff; }

private:
    X& thisX() { return *static_cast<X*>(this); }
    X const& thisX() const { return *static_cast<X const*>(this); }
};

/*
 * Iterator position differences.
 */

template <class X,class S,class D>
X operator+( D diff, IteratorBase<X,S,D> const& itr )
{ return itr+diff; }

template <class T>
class FwdIterator
: public std::iterator<std::random_access_iterator_tag,T>,
  public IteratorBase<FwdIterator<T>,T*,ptrdiff_t>
{
public:
    FwdIterator() {}
    FwdIterator( T* pos ) : IteratorBase<FwdIterator<T>,T*,ptrdiff_t>(pos) {}

    // compiler-supplied copying, destruction are OK

    T& operator*() const { return *this->mPos; }
    T* operator->() const { return this->mPos; }
    T& operator[]( ptrdiff_t diff ) const { return this->mPos[diff]; }
};

template <class T>
class FwdConstIterator
: public std::iterator<std::random_access_iterator_tag,T,ptrdiff_t,T const*,T const&>,
  public IteratorBase<FwdConstIterator<T>,T const*,ptrdiff_t>
{
public:
    FwdConstIterator() {}

    FwdConstIterator( T const* pos )
    : IteratorBase<FwdConstIterator<T>,T const*,ptrdiff_t>(pos) {}

    FwdConstIterator( IteratorPosition<T const*> const& pos )
    : IteratorBase<FwdConstIterator<T>,T const*,ptrdiff_t>(pos.pos()) {}

    FwdConstIterator( IteratorPosition<T*> const& pos )
    : IteratorBase<FwdConstIterator<T>,T const*,ptrdiff_t>(pos.pos()) {}

    // compiler-supplied copying, destruction are OK

    T const& operator*() const { return *this->mPos; }
    T const* operator->() const { return this->mPos; }
    T const& operator[]( ptrdiff_t diff ) const { return this->mPos[diff]; }
};

template <class T>
class RevIterator
: public std::iterator<std::random_access_iterator_tag,T>,
  public IteratorBase<RevIterator<T>,unsigned long,ptrdiff_t>
{
public:
    RevIterator() {}

    RevIterator( unsigned long idx, T* pEnd )
    : IteratorBase<RevIterator<T>,unsigned long,ptrdiff_t>(idx), mpEnd(pEnd) {}

    // compiler-supplied copying, destruction are OK

    T& operator*() const { return mpEnd[-this->mPos-1]; }
    T* operator->() const { return mpEnd-this->mPos-1; }
    T& operator[]( ptrdiff_t diff ) const { return mpEnd[-this->mPos-diff-1]; }

private:
    T* mpEnd;
};

template <class T>
class RevConstIterator
: public std::iterator<std::random_access_iterator_tag,T,ptrdiff_t,T const*,T const&>,
  public IteratorBase<RevConstIterator<T>,unsigned long,ptrdiff_t>
{
public:
    RevConstIterator() {}

    RevConstIterator( unsigned long idx, T const* pEnd )
    : IteratorBase<RevConstIterator<T>,unsigned long,ptrdiff_t>(idx),
      mpEnd(pEnd) {}

    RevConstIterator( IteratorPosition<unsigned long> const& pos )
    : IteratorBase<RevConstIterator<T>,unsigned long,ptrdiff_t>(pos.pos()) {}

    // compiler-supplied copying, destruction are OK

    T const& operator*() const { return mpEnd[-this->mPos-1]; }
    T const* operator->() const { return mpEnd-this->mPos-1; }

    T const& operator[]( ptrdiff_t diff ) const
    { return mpEnd[-this->mPos-diff-1]; }

private:
    T const* mpEnd;
};

#endif /* FEUDAL_ITERATOR_H_ */
