/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////
/*
 * \file QualNibbleVec.h
 * \author iainm
 * \date Sep 1, 2009
 *
 * \brief Quality scores compressed to a half-byte.
 */
#ifndef QUAL_NIBBLE_VECTOR_H
#define QUAL_NIBBLE_VECTOR_H

#include "String.h"
#include "Qualvector.h"
#include "feudal/FieldVec.h"
#include "feudal/MasterVec.h"
#include "feudal/Mempool.h"
#include "system/Assert.h"
#include <algorithm>

class QualNibbleVec : public FieldVec<4, MempoolAllocator<unsigned char> >
{
public:
  typedef allocator_type Alloc;
  typedef FieldVec<4, MempoolAllocator<unsigned char> > BaseT;

  struct QualMapper
  {
      value_type operator()( value_type val ) const
      { return deflate(val); }

      static value_type deflate( value_type val )
      { using std::min; return min(val/3,15); }
      static value_type inflate( value_type val )
      { return 3*val + 1; }
  };

  //
  // Constructors
  //
  QualNibbleVec() {}
  QualNibbleVec( Alloc const& alloc ) : BaseT(alloc) {}

  // SetToSubOf constructor.
  QualNibbleVec( const QualNibbleVec& q, const size_type start, const size_type len )
  { AssertLe(start,q.size()); AssertLe(len,q.size()-start);
    assign(q.begin(start),q.begin(start+len)); }

  // sized constructor
  explicit QualNibbleVec( size_type sz, size_type extra = 0 )
  : BaseT(sz, value_type(), sz+extra) {}

  // Allocator constructor
  QualNibbleVec( QualNibbleVec const& that, Alloc const& alloc )
  : BaseT(that, alloc) {}

  // Copy constructor
  QualNibbleVec( QualNibbleVec const& q )
  : BaseT(q) {}

  // Feudal constructor
  QualNibbleVec( BinaryReader& rdr, size_t varDataLen, void* nElems_unaligned,
                  Alloc const& alloc = Alloc() )
  : BaseT(rdr,varDataLen,nElems_unaligned,alloc) {}

  // Construct from an 8 bit qualvector
  explicit QualNibbleVec( qualvector const & qv )
  { assign(qv.begin(),qv.end(),QualMapper()); }


  // TODO: the iterators for this class will give mapped values, not Phred scores

  //
  // Converstions to and from 8 bit qualvector
  //

  // Initialize QualNibbleVec from an 8 bit qualvector
  void SetFromQualvector( const qualvector& q )
  { assign(q.begin(),q.end(),QualMapper()); }

  // Convert QualNibbleVec to an 8 bit qualvector
  qualvector GetQualvector() const
  { qualvector qv; qv.reserve(size());
    for (size_type i = 0; i < size(); ++i) qv.push_back( (*this)[i] );
    return qv; }


  //
  // Accessors
  //

  unsigned char operator[]( size_type idx ) const
  { return QualMapper::inflate(BaseT::operator[](idx)); }

  void set( size_type idx, unsigned char value)
  { BaseT::set(idx, QualMapper::deflate(value) ); }
};

SELF_SERIALIZABLE(QualNibbleVec);

typedef MasterVec<QualNibbleVec> VecQualNibbleVec;


void WriteAll( const VecQualNibbleVec & quals, const String & outfile );






/* LoadQualNibbleVec: Load a VecQualNibbleVec from a qualb file.
 *
 ******************************************************************************/
void LoadQualNibbleVec( const String & quals_file,
			VecQualNibbleVec * quals,
			const bool CAP_QUALS = false,
			const size_t QUAL_CEIL_RADIUS = 0 );

// Helper function for LoadQualNibbleVec.
template<class Q_t> void CapQualityScores(const int qual_ceil_radius, Q_t * quals);







#endif
