///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2009) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////





// --------------------------------------
// MapOfCounters
// --------------------------------------
class MapOfCounters : public map<size_t, size_t>
{
public:
  MapOfCounters() : map<size_t, size_t>() {}

  template <class T>
  MapOfCounters(const vec<T> & v) 
    : map<size_t, size_t>()
  {
    for (typename vec<T>::const_iterator it = v.begin(); it != v.end(); it++)
      (*this)[*it]++;
  }


  size_t Total() const 
  {
    size_t total = 0;
    for (MapOfCounters::const_iterator it = (*this).begin(); 
         it != (*this).end(); it++)
      total += it->second;
    return total;
  }

  void Write(const String & filename) const
  {
    double total = static_cast<double>(Total());

    ofstream ofs;
    ofs.open(filename.c_str());

    ofs << "# 1:x 2:y 3:y/total 4:cum(y) 5:cum(y)/total" << endl;

    size_t cum = 0;
    for (MapOfCounters::const_iterator it = (*this).begin(); 
       it != (*this).end(); it++) {
      cum += it->second;
      ofs << " " << it->first
          << " " << it->second
          << " " << setprecision(10) << static_cast<double>(it->second) / total
          << " " << cum
          << " " << setprecision(10) << static_cast<double>(cum) / total
          << endl;
    }
    ofs.close();
  }


  MapOfCounters & operator += (const MapOfCounters & m) 
  {
    for (MapOfCounters::const_iterator it = m.begin(); it != m.end(); it++)
      (*this)[(*it).first] += (*it).second;
    return *this;
  } 

};




// --------------------------------------
// KmerFrequencyCounters
// --------------------------------------
class KmerFrequencyCounters : public MapOfCounters 
{
public:
  void Write(const String & filename) const
  {
    // ---- counting kmers and distinct kmers
    const size_t total_num_distinct_kmers = (*this).Total();
  
    size_t total_num_kmers = 0;
    for (KmerFrequencyCounters::const_iterator it = (*this).begin(); 
         it != (*this).end(); it++) {
      
      const size_t & kmer_frequency = it->first;
      const size_t & num_distinct_kmers = it->second;
      
      total_num_kmers += kmer_frequency * num_distinct_kmers;
    }
    
    // ---- writing kmer frequency stats to disk
    ofstream outfs;
    outfs.open(filename.c_str());
    outfs << ("#"
              " 1:kmer_frequency"
              " 2:num_distinct_kmers"
              " 3:2/total_num_distinct_kmers"
              " 4:cummulative(2)"
              " 5:4/total_num_distinct_kmers"
              " 6:1*2"
              " 7:6/total_num_kmers"
              " 8:cummulative(6)"
              " 9:8/total_num_kmers") 
          << endl;


    size_t cum_num_distinct_kmers = 0;
    size_t cum_num_kmers = 0;
    for (KmerFrequencyCounters::const_iterator it = (*this).begin(); 
         it != (*this).end(); it++) {
    
      const size_t & kmer_frequency = it->first;
      const size_t & num_distinct_kmers = it->second;

      cum_num_distinct_kmers += num_distinct_kmers;
      const size_t num_kmers = kmer_frequency * num_distinct_kmers;
      cum_num_kmers += num_kmers;
    
      outfs << " " << kmer_frequency 
            << " " << num_distinct_kmers
            << " " << setprecision(10) << static_cast<double>(num_distinct_kmers) / static_cast<double>(total_num_distinct_kmers) 
            << " " << cum_num_distinct_kmers 
            << " " << setprecision(10) << static_cast<double>(cum_num_distinct_kmers) / static_cast<double>(total_num_distinct_kmers) 
            << " " << num_kmers
            << " " << setprecision(10) << static_cast<double>(num_kmers) / static_cast<double>(total_num_kmers) 
            << " " << cum_num_kmers 
            << " " << setprecision(10) << static_cast<double>(cum_num_kmers) / static_cast<double>(total_num_kmers) 
            << endl;
    }
  
    outfs.close();
  }

};




// --------------------------------------
// ComparativeKmerFrequencyCounters
// --------------------------------------
class ComparativeKmerFrequencyCounters
{
public:
  KmerFrequencyCounters both1, both2, only1, only2;

  ComparativeKmerFrequencyCounters() {}

  ComparativeKmerFrequencyCounters & operator += (const ComparativeKmerFrequencyCounters & c) 
  {
    this->both1 += c.both1;
    this->both2 += c.both2;
    this->only1 += c.only1;
    this->only2 += c.only2;
    return *this;
  } 
};




// --------------------------------------
// KmerFrequencyStatistics
// --------------------------------------

class KmerFrequencyStatistics
{
private: 
  LockedData mutexes_min[512];
  LockedData mutexes_max[512];

  vec<KmerFrequencyCounters> kmer_freq_counters;
  bool collapsed;

public:
  vec<uint32_t> reads_min_kmer_freq;
  vec<uint32_t> reads_max_kmer_freq;

  KmerFrequencyStatistics(const size_t num_reads, 
                          const size_t num_blocks) 
  {
    reads_min_kmer_freq.resize(num_reads, 1 << 30);
    reads_max_kmer_freq.resize(num_reads, 0);

    kmer_freq_counters.resize(num_blocks);
  }


  void UpdateMinMax(const size_t read_ID, 
                    const size_t kmer_freq)
  {
    if (true) {
      Locker lock(mutexes_max[read_ID & 0x1FF]);
      if (reads_max_kmer_freq[read_ID] < kmer_freq) 
        reads_max_kmer_freq[read_ID] = kmer_freq;
    }
    if (true) {
      Locker lock(mutexes_min[read_ID % 0x1FF]);
      if (reads_min_kmer_freq[read_ID] > kmer_freq)
        reads_min_kmer_freq[read_ID] = kmer_freq;
    }
  }

  void UpdateFreqs(const size_t block_ID, const size_t kmer_freq)
  {
    kmer_freq_counters[block_ID][kmer_freq]++;
    collapsed = false;
  }


  void Update(const vec<const KmerLoc*> & kmer_locs_p,
              const size_t block_ID)
  {
    size_t kmer_freq = kmer_locs_p.size();
    
    for (vec<const KmerLoc*>::const_iterator it = kmer_locs_p.begin(); it != kmer_locs_p.end(); ++it) 
      UpdateMinMax((*it)->GetReadID(), kmer_freq);
    UpdateFreqs(block_ID, kmer_freq);
  }


  KmerFrequencyCounters & AllKmerFrequencyCounters()
  {
    if (!collapsed) {
      // accumulate all freqs on kmer_freq_counters[0]
      
      for (size_t ib = 1; ib != kmer_freq_counters.size(); ++ib) {
        kmer_freq_counters[0] += kmer_freq_counters[ib];       
        kmer_freq_counters[ib].clear();
      }
      collapsed = true;
    }  
    return kmer_freq_counters[0];
  }

};


