/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////

#ifndef FORCE_DEBUG
    #define NDEBUG
#endif

#include <map>

#include "Basevector.h"
#include "Bitvector.h"
#include "CoreTools.h"
#include "FeudalMimic.h"
#include "kmers/KmerRecord.h"
#include "kmers/SortKmers.h"
#include "pairwise_aligners/BalancedMutmerGraph.h"
#include "paths/KmerPath.h"
#include "paths/MakeAlignsPathsParallelX.h"
#include "paths/ReadsToPathsCoreX.h"

static inline 
String Tag(String S = "RTPCX") { return Date() + " (" + S + "): "; } 

// Returns true if the K-mer at basevector b, in the range [offset, offset+K),
// is palindromic.
Bool isPalindrome( const basevector & b, const int K, unsigned int offset )
{
  if ( K % 2 ) return false; // kmers cannot be palindromic if K is odd
  for ( int i = 0; i < K/2; i++ )
    if ( b[offset+i] + b[offset+K-i-1] != 3 ) return false;
  return true;
}



class MHit {
private:

  // Total memory usage: 23 bytes (stored as 24)
  int64_t id1_;     // long long
  int64_t id2_;
  uint16_t pos1_;   // unsigned short
  uint16_t pos2_;
  uint16_t len_;
  Bool rc_;

public:

  MHit() { }
  MHit(int64_t id1, 
       int64_t id2, 
       Bool rc,
       uint16_t pos1, 
       uint16_t pos2, 
       uint16_t len)
    : id1_(id1), id2_(id2), pos1_(pos1), pos2_(pos2), len_(len), rc_(rc) { }

  void Set(int64_t id1, 
           int64_t id2, 
           Bool rc,
           uint16_t pos1, 
           uint16_t pos2, 
           uint16_t len) {
    id1_ = id1;
    id2_ = id2;
    pos1_ = pos1;
    pos2_ = pos2;
    len_ = len;
    rc_ = rc;
  }

  inline int64_t Id1() const { return id1_; }
  inline int64_t Id2() const { return id2_; }
  inline uint16_t pos1() const { return pos1_; }
  inline uint16_t pos2() const { return pos2_; }
  inline uint16_t Len() const { return len_; }
  inline Bool Rc2() const { return rc_; }

  inline void SetId1(int64_t id1) { id1_ = id1; }
  inline void SetId2(int64_t id2) { id2_ = id2; }

  friend Bool operator<(const MHit& h1, const MHit& h2) {
    if (h1.Id2() < h2.Id2()) return True;
    if (h1.Id2() > h2.Id2()) return False;
    return h1.Len() > h2.Len();    
  }
  
  friend std::istream& operator>>(std::istream & is, MHit & mhit) 
  { is.read(reinterpret_cast<char*>(&mhit), sizeof(MHit)); return is; }

  friend std::ostream& operator<<(std::ostream & os, const MHit & mhit)
  { os.write(reinterpret_cast<const char*>(&mhit), sizeof(MHit)); return os; }
   

};



// NOTE: this is NOT longlong-safe
template<int I, int K> 
void MakeAlignsPathsX(const vecbasevector& EE, 
                      vecbitvector& chosen,
                      BMG<I> & bmg, 
                      const bool OPTIM_MEM )
{
  longlong N = EE.size( );
  vec<int> rid(N, vec<int>::IDENTITY );
  longlong S_init = 0;
  for ( size_t l = 0; l < EE.size( ); l++ )
    if ( EE[l].size( ) >= K ) S_init += EE[l].size( ) - K;
  S_init += S_init/4;
  S_init /= 33;

  // Balanced Mutmer Graph: this object takes up a LOT of memory
  bmg.resize(N);

  // If OPTIM_MEM, optimize for memory by using SortKmers<100>;
  // otherwise, optimize for runtime by using SortKmers<10>
  int n_passes = OPTIM_MEM ? 100 : 10;
  for ( int pass = 0; pass < n_passes; pass++ ) {
       
    unsigned int S = S_init;
    vec< kmer_record<K,I> > R(S);
       
    if ( OPTIM_MEM ) {
      dummy<100> SK_dummy;
      SortKmers( SK_dummy, EE, rid, pass, R, S );
    }
    else {
      dummy<10> SK_dummy;
      SortKmers( SK_dummy, EE, rid, pass, R, S );
    }
       
    ForceAssertGe( R.size( ), S );
    //cout << Date( ) << ": Pass " << pass << ": S = " << S << endl;
       
    // Loop through the KmerRecords as returned by SortKmers.
    longlong jj = 0, r2 = 0;
    longlong read_id1 = -1;
    int pos1 = -1;
    while(1) {
	 
      if ( r2 + 1 < jj ) {
        r2++;
      }
      else {
	longlong i = jj;
        while ( 1 ) {
	     
          // Check if we're out of KmerRecords.
          if ( i == S ) {
            read_id1 = -1;
            break;
          }
	     
          // Get batch of k-mer records corresponding to one k-mer.
          int l;
          for ( jj = i+1; jj < S; jj++ ) {
            for ( l = (K+3)/4 - 1; l >= 0; l-- )
              if ( R[jj].Bytes( )[l] != R[i].Bytes( )[l] )
                break;
            if ( l >= 0 ) break;
          }
          // Sort this batch.
          sort( R.begin( ) + i, R.begin( ) + jj, kmer_record<K,I>::id_cmp_pos );
          read_id1 = R[i].GetId( );
          pos1 = R[i].GetPos( );
	     
          // Mark canonical placement of k-mer.
          int xpos1 = pos1;
          if ( pos1 < 0 ) xpos1 = -pos1;
          --xpos1;
          chosen[read_id1].Set( xpos1, True );
	     
          if ( i + 1 >= jj ) { i = jj; continue; } // only one kmer in batch
	     
          // Go through pairs.
          r2 = i + 1;
          break;
        }
      }
      if ( read_id1 < 0 ) break;
	 
      // Merge into mutmer graph.
	 
      longlong read_id2 = R[r2].GetId( );
      int pos2 = R[r2].GetPos( );
      bmg.MergeKmer( pos1, pos2, K, read_id1, read_id2, 
                     EE[read_id1], EE[read_id2], True );
    }
  }
}





template<int I>
void GenerateMutmersX(BMG<I>& bmg, 
                      const vecbasevector & EE,
		      vec<MHit> & mhits)
{
  size_t num_ids = EE.size();
  //  cout << Tag() << " (GMsX): num_ids = " << num_ids << endl;
  //cout << Tag() << " (GMsX): bmg.size = " << bmg.size() << endl;
  //cout << Tag() << " (GMsX): bmg.SizeOf = " << bmg.SizeOf() << endl;
  
  if ( bmg.size() == 0 ) return;

  // Generate mutmers.
  for ( size_t id1 = 0; id1 < num_ids; id1++ ) {

    BMG_node<I> & mmids = bmg[id1];

    size_t count = mmids.size();

    sort( mmids.begin(), mmids.begin() + count );
    for ( size_t im0 = 0, im1 = 1; im0 < count; im0 = im1 ) {

      while (im1 < count && mmids[im0] == mmids[im1]) im1++;

      bool RC = mmids[im0].Rc();
      size_t id2 = mmids[im0].UnpackReadId();
      for ( size_t im = im0; im != im1; im++ ) {
        uint16_t pos1, pos2, len;
        mmids[im].Unpack( pos1, pos2, len );
        MHit h;
        if ( id1 < id2 ) h.Set( id1, id2, RC, pos1, pos2, len );
        else if ( !RC )  h.Set( id2, id1, RC, pos2, pos1, len );
        else {
          size_t n1 = EE[id1].size();
          size_t n2 = EE[id2].size();
          h.Set( id2, id1, RC, n2-pos2-len, n1-pos1-len, len );    
        }
        mhits.push_back(h);
      }
    }

    mmids.clear();
  }

  //cout << Tag() << " (GMsX): mhits.size = " << mhits.size() << endl; 
}


bool MHitsOnFile(const String CHECKPOINT_HEAD)
{
  return IsRegularFile(CHECKPOINT_HEAD + ".mhits");
}

void MHitsToFile(const vec<MHit> & mhits, const String CHECKPOINT_HEAD) 
{
  ofstream os;
  os.open((CHECKPOINT_HEAD + ".mhits").c_str());
  int64_t n_mhits = mhits.size();
  os.write(reinterpret_cast<const char*>(&(n_mhits)), sizeof(int64_t));
  for (int64_t i = 0; i != n_mhits; i++)
    os << mhits[i];
  os.close();
}


void MHitsFromFile(vec<MHit> & mhits, const String CHECKPOINT_HEAD)
{
  String chkpt_fn = CHECKPOINT_HEAD + ".mhits";
 
  ifstream is;
  is.open((CHECKPOINT_HEAD + ".mhits").c_str());
  int64_t n_mhits;
  is.read(reinterpret_cast<char*>(&n_mhits), sizeof(int64_t));
  mhits.resize(n_mhits);
  for (int64_t i = 0; i != n_mhits; i++)
    is >> mhits[i];
  is.close();
}




void ReadsToPathsMainPartX(const longlong        N, 
                           vec<longlong>       & new_order, 
                           const int             K, 
                           const vecbasevector & EE, 
                           vecbasevector       & EEX, 
                           const bool            big_BMG, 
                           const String        & PARCEL_DIR,
                           const int             NUM_THREADS,
                           const String          CHECKPOINT_HEAD,
                           vecKmerPath         & pathsv)
{
  // Call MakeAligns with the goal of creating a BMG (Balanced Mutmer Graph)
  // The BMG has a parameter I, indicating internal storage space
  // If the the reads are long (>1024bp), we must use I=2;
  // otherwise, we can use I=1, which is more efficient
  // This is the big runtime/memory bottleneck of pathing!

#define CALL_MAKE_ALIGNS( I, __K )                           \
  {	                                                     \
    BMG<I> bmg;                                              \
    if (1)						     \
      MakeAlignsPathsParallelX<I>                            \
        (__K, EEX, chosen, bmg,                              \
         PARCEL_DIR, NUM_THREADS, CHECKPOINT_HEAD);          \
    else                                                     \
      MakeAlignsPathsX<I, __K>(EEX, chosen, bmg, True);      \
                                                             \
    GenerateMutmersX<I>(bmg, EEX, mhits);                    \
  }
     
#define CASE(_K)                                  \
  {                                               \
    if (big_BMG)  { CALL_MAKE_ALIGNS( 2, _K );	} \
    else          { CALL_MAKE_ALIGNS( 1, _K );  } \
  }
     

  vec<MHit> mhits;
  vecbitvector chosen;

  if ( 0 &&    // for now, this never happens
       CHECKPOINT_HEAD != "" && 
       MHitsOnFile(CHECKPOINT_HEAD)) {
      
    cout << Tag("RTPMPX") << "CHECKPOINT: retrieving previously computed 'mhits'." << endl;
    MHitsFromFile(mhits, CHECKPOINT_HEAD);
    cout << Tag("RTPMPX") << "CHECKPOINT: retrieved " << mhits.size() << " mhits." << endl;

  } 
  else {
    //cout << Tag("RTPMPX") << "computing 'mhits' and 'chosen'." << endl;
    Mimic(EEX, chosen);

    DISPATCH_ON_K(K, CASE);
    
    if (CHECKPOINT_HEAD != "") {
      cout << Tag("RTPMPX") << "CHECKPOINT: saving 'mhits'." << endl;
      MHitsToFile(mhits, CHECKPOINT_HEAD);
      cout << Tag("RTPMPX") << "CHECKPOINT: saved " << mhits.size() << " mhits." << endl;
    }     
  }

  Destroy(EEX);

  //cout << Tag("RTPMPX") << "Reorder reads" << endl;

  // Re-order the reads, undoing the sorting by length that we did earlier.
  longlong n_mhits = mhits.size( );
  {
    for ( longlong i = 0; i < n_mhits; i++ ) {
      mhits[i].SetId1( new_order[ mhits[i].Id1( ) ] );
      mhits[i].SetId2( new_order[ mhits[i].Id2( ) ] );    
    }
    vecbitvector chosen2;
    Mimic( EE, chosen2 );
    for ( longlong i = 0; i < N; i++ )
      chosen2[ new_order[i] ] = chosen[i];
    chosen = chosen2;    
    Destroy( new_order );
  }
  
  //cout << Tag("RTPMPX") << "Sort mhits" << endl;
  // Sort the mhits.
  Sort(mhits);

     

  // Find which canonical kmers (i.e., kmers marked as chosen) are
  // palindromic.

  //cout << Tag("RTPMPX") << "Finding palindromes" << endl;
  vecbitvector chosen_palindrome;
  Mimic( EE, chosen_palindrome );
  for ( size_t i = 0; i < chosen.size( ); i++ )
    for ( unsigned int j = 0; j < chosen[i].size( ); j++ )
      if ( chosen[i][j] )
	if ( isPalindrome( EE[i], K, j ) )
	  chosen_palindrome[i].Set( j, True );

  // Number positions on the reads.

  size_t kx = 0;
  vec<longlong> start( EE.size( ) );
  for ( size_t i = 0; i < EE.size( ); i++ ) {
    start[i] = kx;
    kx += EE[i].size( );    
  }
  ForceAssertEq( kx, EE.sumSizes() );

  // Build read paths for the reads themselves.
  //cout << Tag("RTPMPX") << "Building read paths" << endl;

  vec<KmerPath> paths;
  paths.reserve(N);
  vec<longlong> index;
     
  // Set up tracking of palindromes.

  longlong npalindromes = 0;
  map<longlong,longlong> palindromic_id;

  longlong nsegments = 0;
  longlong i = 0;
  // Loop over all reads.  For each read, fill an index vector that indicates
  // what the kmer_ids on that read will be.
  for ( longlong id2 = 0; id2 < N; id2++ ) {
    longlong j;
    for ( j = i; j < n_mhits; j++ )
      if ( mhits[j].Id2( ) != id2 ) break;
    unsigned short n2 = EE[id2].size( );
    index.resize_and_set( Max( 0, n2 -  K + 1 ), -1 );
    // Find canonical kmers on this read (there may be any number, including 0)
    // and mark them with a kmer_id corresponding to their location.
    for ( int u = 0; u <= n2 -  K; u++ )
      if ( chosen[id2][u] ) index[u] = halfway_kmer + start[id2] + u + 1;
       
    for ( longlong k = i; k < j; k++ ) {
      longlong id1 = mhits[k].Id1( );
      unsigned short pos1 = mhits[k].pos1( ), pos2 = mhits[k].pos2( );
      unsigned short len = mhits[k].Len( );
      Bool rc2 = mhits[k].Rc2( );

      // Validate mutmer.
      // Each mutmer hit (mhit) represents a kmer equivalance: it claims that
      // the bases of read #id1, starting at pos1, are equal to (or rc of)
      // the bases of read #id2, starting at pos2, for a length of len bases.
      // If these asserts fail, the bases are not actually equal.
      if ( !rc2 ) {
        for ( unsigned short u = 0; u < len; u++ )
	  ForceAssertEq( EE[id1][pos1+u], EE[id2][pos2+u] );
      }
      else
	for ( unsigned short u = 0; u < len; u++ )
	  ForceAssertEq( 3-EE[id1][pos1+u], EE[id2][n2-pos2-u-1] );

      // Look through all the mutmer hits involving this read, in search of
      // correspondences between a kmer on this read and its canonical version
      // on other reads.  For all such correspondences found, mark the location
      //  on this read with the same kmer_id used by the canonical kmer on the
      // other read.
      if ( !rc2 ) {
        for ( unsigned short u = 0; u <= len -  K; u++ ) {
          if ( !chosen[id1][pos1+u] ) continue;
	  if ( index[pos2+u] >= 0 ) continue;
          index[pos2+u] = halfway_kmer + start[id1] + pos1 + u + 1;
        }
      }
      else {
        for ( unsigned short u =  K - 1; u < len; u++ ) {
          unsigned short up = u -  K + 1;
          if ( !chosen[id1][ pos1 + up ] ) continue;
          if ( index[n2-pos2-u-1] >= 0 ) continue;
          if ( chosen_palindrome[id1][ pos1 + up ] )
            index[n2-pos2-u-1] = halfway_kmer + start[id1] + pos1 + up + 1;
          else
            index[n2-pos2-u-1] = halfway_kmer - start[id1] - pos1 - up;
        }
      }
    }

    // Check for palindromes.
    for ( size_t u = 0; u < index.size( ); u++ ) {
      if ( !isPalindrome( EE[id2], K, u ) ) continue;
      longlong pid = palindromic_id[ index[u] ];
      if ( pid == 0 ) {
	ForceAssertLt( npalindromes, max_palindromes );
	pid = first_palindrome + npalindromes;
	palindromic_id[ index[u] ] = pid;
	++npalindromes;    
      }
      index[u] = pid;
    }

    // Build the new KmerPath, segment by segment.
    KmerPath p;
    for ( int u = 0; u <= n2 -  K; u++ ) {
      int v;
      for ( v = u + 1; v <= n2 -  K; v++ )
        if ( index[v] != index[v-1] + 1 ) break;
      ++nsegments;
      p.AddSegment( index[u], index[u] + v - u - 1 );
      u = v - 1;    
    }
    paths.push_back(p);
       
    // Sanity check.  If this assert fails, the index vector was not built correctly.
    // Most likely, there are missing mutmer hits, and some base-locations are
    // not marked with kmer IDs as they should have been.
    for ( int u = 0; u <= n2 -  K; u++ ) {
      if ( index[u] < 0 ) PRINT3_TO( cout, id2, u, n2 );
      ForceAssert( index[u] >= 0 );
    }
    i = j;    
  }
  
  Destroy( mhits );
  Destroy( start );
  chosen.clear().shrink_to_fit();
  chosen_palindrome.clear().shrink_to_fit();
  palindromic_id.clear();
     
  // Convert from vec<KmerPath> to vecKmerPath.
  pathsv = FeudalOfKmerPath( paths );
}


void ReadsToPathsCoreX(const vecbasevector & EE, 
                       const int             K, 
                       vecKmerPath         & paths, 
                       const String        & PARCEL_DIR,
                       const int             NUM_THREADS,
                       const String          CHECKPOINT_HEAD)
{
  ForceAssertSupportedK( K );

  // Determine whether this dataset will require a big BalancedMutmerGraph
  // (e.g., one with I=2 rather than I=1.)
  longlong N = EE.size( );
  bool big_BMG = false;
  for ( longlong i = 0; i < N; i++ ) {
    if ( EE[i].size( ) >= 1024 ) big_BMG = True;
    ForceAssertLe( EE[i].size( ), static_cast<unsigned>(USHRT_MAX) ); // ReadsToPathsCoreY should ensure that this is true
  }
  if ( N > INT_MAX ) big_BMG = true;
  ForceAssertLe( N, longlong(1) << 47 );

  if ( EE.sumSizes() > static_cast<size_t>(halfway_kmer) )
    FatalErr( "Five byte limit on k-mer indices exceeded." );
  
  // Place reads in descending order by size.  The vector new_order remembers the 
  // new order, which is reversed later.
  vec<longlong> new_order(N);
  vecbasevector EEX = EE;
  {
    vec< pair<int,longlong> > lengths(N);
    for ( longlong i = 0; i < N; i++ )
      lengths[i] = make_pair( EE[i].size( ), i );
    ReverseSort(lengths);
    for ( longlong i = 0; i < N; i++ )
      new_order[i] = lengths[i].second;
       
       
    vec<longlong> new_order2 = new_order, new_map(N);
    for ( longlong i = 0; i < N; i++ )
      new_map[ new_order2[i] ] = i;
    for ( longlong i = 0; i < N; i++ ) {
      if ( i != new_order2[i] ) {
        swap( EEX[i], EEX[ new_order2[i] ] );
        longlong saved_new_order2_i = new_order2[i];
        swap( new_order2[i], new_order2[ new_map[i] ] );
        swap( new_map[i], new_map[saved_new_order2_i] );
      }
    }
  }

  // Go through one pass.
  ReadsToPathsMainPartX(N, new_order, K, EE, EEX, big_BMG,
			PARCEL_DIR, NUM_THREADS, CHECKPOINT_HEAD, paths);
}

// ReadsToPathsCoreX will fail if it faces the prospect of a KmerPathInterval
// with length greater than 65535.
// ReadsToPathsCoreY is a wrapper which solves this:
// it breaks long kmer paths, call ReadsToPathsCoreX, 
// and puts broken things back together.

void ReadsToPathsCoreY(const vecbasevector & EE, 
                       const int             K, 
                       vecKmerPath         & paths, 
                       const String        & PARCEL_DIR,
                       const int             NUM_THREADS,
                       const String          CHECKPOINT_HEAD)
{
  // Are there any long reads?
  // "Long": greater than USHRT_MAX, the maximum unsigned short, which is 65535.
  bool long_read = false;
  for(size_t i=0; i < EE.size() && !long_read; i++)
    long_read = ( EE[i].size() > USHRT_MAX );
 
  // If not, Y=X.
  if( ! long_read ) {
    ReadsToPathsCoreX(EE, K, paths, PARCEL_DIR, NUM_THREADS, CHECKPOINT_HEAD);
  }
  else {

    // If so, we need to break up the reads.
    longlong nreads = 0;
    longlong reads_rawsize = 0;
    vec<size_t> contig;
    vecbasevector sheared_reads;
    basevector b;

    for ( int pass = 1; pass <= 2; pass++ ) {
      if ( pass == 2 ) {
        sheared_reads.Reserve( reads_rawsize, nreads );
        contig.reserve(nreads);
      }
      for ( size_t i = 0; i < EE.size( ); i++ ) {
        for ( size_t start = 0; start < EE[i].size(); start++ ) {
	  size_t len = EE[i].size() - start;
	  if ( len > USHRT_MAX ) len = USHRT_MAX;
	  
          if ( pass == 1 ) {
            nreads++;
            reads_rawsize += (len+15)/16;
          }
          else {
            b.SetToSubOf( EE[i], start, len );
            sheared_reads.push_back(b);
            contig.push_back(i);
          }
	  
          if ( start + len == EE[i].size() ) break;
          start += len - K;
        }
      }
    }

    // Now the vecbasevector sheared_reads holds the sheared version of EE.
    vecKmerPath sheared_paths;
    ReadsToPathsCoreX(sheared_reads, K, sheared_paths, PARCEL_DIR, NUM_THREADS, CHECKPOINT_HEAD);

    // Now release the memory that held the sheared reads...
    Destroy(sheared_reads);

    // Finally, reassemble.  We have to do this gently,
    //  so that nothing ends up self-owned.
    paths.clear().reserve( EE.size() );

    KmerPath one_path;
    size_t which_shear = 0;

    for( size_t which_path = 0; which_path < EE.size(); which_path++ ) {
      one_path.Clear();
      while( which_shear < contig.size() && 
             contig[which_shear] == which_path ) {
        one_path.Append( sheared_paths[which_shear] );
        which_shear++;
      }
      paths.push_back( one_path );
    }
  }
}


// Wrapper functions, which also create paths_rc and/or pathsdb.

void ReadsToPathsCoreY(const vecbasevector & EE, 
                       const int             K,
                       vecKmerPath         & paths, 
                       vecKmerPath         & paths_rc, 
                       vec<tagged_rpint>   & pathsdb,
                       const String        & PARCEL_DIR,
                       const int             NUM_THREADS,
                       const String          CHECKPOINT_HEAD)
{
  ReadsToPathsCoreY(EE, K, paths, PARCEL_DIR, NUM_THREADS, CHECKPOINT_HEAD);
  paths_rc = paths;
  for ( size_t i = 0; i < paths_rc.size( ); i++ )
    paths_rc[i].Reverse( );
  CreateDatabase( paths, paths_rc, pathsdb );    
}
  


void ReadsToPathsCoreY(const vecbasevector   & EE, 
                       const int               K,
                       vecKmerPath           & paths, 
                       vecKmerPath           & paths_rc, 
                       vec<big_tagged_rpint> & pathsdb,
                       const String          & PARCEL_DIR,
                       const int               NUM_THREADS,
                       const String            CHECKPOINT_HEAD)
{ 
  ReadsToPathsCoreY(EE, K, paths, PARCEL_DIR, NUM_THREADS, CHECKPOINT_HEAD);
  paths_rc = paths;
  for (size_t i = 0; i < paths_rc.size( ); i++)
    paths_rc[i].Reverse( );
  CreateDatabase( paths, paths_rc, pathsdb );    
}

