///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////
/*
 * \file SortInPlace.h
 * \author tsharpe
 * \date Feb 23, 2010
 *
 * \brief
 */
#ifndef SORTINPLACE_H_
#define SORTINPLACE_H_

#include "Compare.h"
#include "system/SysConf.h"
#include "system/Worklist.h"
#include <algorithm>
#include <utility>
#include <unistd.h>

template <class Itr, class Comp>
class InPlaceSorter
{
public:
    typedef typename std::iterator_traits<Itr>::difference_type diff_t;
    typedef typename std::iterator_traits<Itr>::value_type value_t;

    InPlaceSorter( Itr const& first, Itr const& last,
                    Comp const& comp = Comp() )
    : mFirst(first), mLast(last), mComp(comp) {}

    void sort()
    { internalSort(mFirst,mLast); }

private:
    Itr iter_median( Itr a, Itr b, Itr c )
    {
        int order = mComp(*a,*b);
        int order2 = mComp(*b,*c);
        if ( !order || !order2 ) return b;
        if ( order < 0 )
        {
            if ( order2 <= 0 ) return b;
            return mComp(*a,*c) <= 0 ? c : a;
        }
        else
        {
            if ( order2 >= 0 ) return b;
            return mComp(*a,*c) <= 0 ? a : c;
        }
    }

    void internalSort( Itr first, Itr last )
    {
        using std::iter_swap;
        while ( true )
        {
            diff_t siz = last - first;
            if ( siz <= INSERTION_SORT_MAX )
            {
                if ( siz > 1 )
                {
                    for ( Itr itr1(first+1); itr1 != last; ++itr1 )
                    {
                        Itr itr2(itr1);
                        while ( itr2 != first )
                        {
                            Itr itr3(itr2--);
                            if ( mComp(*itr3,*itr2) >= 0 )
                                break;

                            iter_swap(itr3,itr2);
                        }
                    }
                }
                break;
            }

            Itr pivot(iter_median(first,first+siz/2,last-1));
            Itr lt(first);
            Itr gt(last);
            Itr itr(first);
            while ( itr != gt )
            {
                int order = mComp(*itr,*pivot);
                if ( order < 0 )
                {
                    iter_swap(lt,itr);
                    if ( pivot == lt )
                        pivot = itr;
                    ++lt;
                    ++itr;
                }
                else if ( order > 0 )
                {
                    iter_swap(itr,--gt);
                    if ( pivot == gt )
                        pivot = itr;
                }
                else
                {
                    ++itr;
                }
            }

            if ( lt-first > 1 )
                internalSort(first,lt);

            first = gt;
        }
    }

    Itr mFirst;
    Itr mLast;
    Comp mComp;

    static diff_t const INSERTION_SORT_MAX = 8;
};

template <class Itr, class Comp>
class InPlaceParallelSorter
{
public:
    typedef typename std::iterator_traits<Itr>::difference_type diff_t;
    typedef typename std::iterator_traits<Itr>::value_type value_t;

    InPlaceParallelSorter( Itr const& first, Itr const& last,
                            Comp const& comp = Comp() )
    : mFirst(first), mLast(last), mComp(comp), mDoneReporter(mLast-mFirst) {}

    void sort()
    {
        if ( mLast-mFirst > 1 )
        {
            getWorklist().add(Workitem(mFirst,mLast,this));
            mDoneReporter.wait();
        }
    }

private:
    Itr iter_median( Itr a, Itr b, Itr c )
    {
        int order = mComp(*a,*b);
        int order2 = mComp(*b,*c);
        if ( !order || !order2 ) return b;
        if ( order < 0 )
        {
            if ( order2 <= 0 ) return b;
            return mComp(*a,*c) <= 0 ? c : a;
        }
        else
        {
            if ( order2 >= 0 ) return b;
            return mComp(*a,*c) <= 0 ? a : c;
        }
    }

    void internalSort( Itr first, Itr last )
    {
        using std::iter_swap;
        size_t done = 0;
        while ( true )
        {
            diff_t siz = last - first;
            if ( siz <= INSERTION_SORT_MAX )
            {
                if ( siz > 1 )
                {
                    for ( Itr itr1(first+1); itr1 != last; ++itr1 )
                    {
                        Itr itr2(itr1);
                        while ( itr2 != first )
                        {
                            Itr itr3(itr2--);
                            if ( mComp(*itr3,*itr2) >= 0 )
                                break;

                            iter_swap(itr3,itr2);
                        }
                    }
                }
                done += siz;
                break;
            }

            Itr pivot(iter_median(first,first+siz/2,last-1));
            Itr lt(first);
            Itr gt(last);
            Itr itr(first);
            while ( itr != gt )
            {
                int order = mComp(*itr,*pivot);
                if ( order < 0 )
                {
                    iter_swap(lt,itr);
                    if ( pivot == lt )
                        pivot = itr;
                    ++lt;
                    ++itr;
                }
                else if ( order > 0 )
                {
                    iter_swap(itr,--gt);
                    if ( pivot == gt )
                        pivot = itr;
                }
                else
                {
                    ++itr;
                }
            }

            done += gt - lt;
            diff_t siz1 = lt - first;
            diff_t siz2 = last - gt;
            if ( siz1 < siz2 )
            {
                if ( siz1 >= PARALLEL_SORT_MIN )
                {
                    getWorklist().add(Workitem(first,lt,this));
                    siz1 = 0;
                }
            }
            else if ( siz2 >= PARALLEL_SORT_MIN )
            {
                getWorklist().add(Workitem(gt,last,this));
                siz2 = 0;
            }

            if ( siz1 <= 1 )
                done += siz1;
            else
                internalSort(first,lt);

            if ( siz2 <= 1 )
            {
                done += siz2;
                break;
            }

            first = gt;
        }

        mDoneReporter.done(done);
    }

    struct DoneReporter : public LockedData
    {
        DoneReporter( size_t toDo ) : mToDo(toDo), mDone(0), mCondVar(*this) {}

        size_t progress()
        { Locker locker(*this); return mDone; }

        void done( size_t siz )
        { Locker locker(*this);
          if ( (mDone += siz) >= mToDo )
              mCondVar.signal(); }

        void wait()
        { Locker locker(*this);
          if ( mDone < mToDo ) locker.wait(mCondVar); }

        size_t mToDo;
        size_t mDone;
        Condition mCondVar;
    };

    struct Workitem
    {
        Workitem() : mpSorter(0) {}

        Workitem( Itr const& first, Itr const& last,
                  InPlaceParallelSorter* pSorter )
        : mFirst(first), mLast(last), mpSorter(pSorter) {}

        Itr mFirst;
        Itr mLast;
        InPlaceParallelSorter* mpSorter;
    };

    struct Processor
    {
        void operator()( Workitem const& wi ) const
        {
            wi.mpSorter->internalSort(wi.mFirst,wi.mLast);
        }
    };
    friend class Processor;

    Itr mFirst;
    Itr mLast;
    Comp mComp;
    DoneReporter mDoneReporter;

    static Worklist<Workitem,Processor>& getWorklist()
    {
        static Worklist<Workitem,Processor> gWorklist(Processor(),
                                                      processorsOnline());
        return gWorklist;
    }

    static diff_t const INSERTION_SORT_MAX = 8;
    static diff_t const PARALLEL_SORT_MIN = 1000;
};

template <class Itr, class Comp>
void sortInPlace( Itr const& first, Itr const& last, Comp const& comp )
{
    InPlaceSorter<Itr,Comp> sorter(first,last,comp);
    sorter.sort();
}

template <class Itr>
void sortInPlace( Itr const& first, Itr const& last )
{
    typedef typename std::iterator_traits<Itr>::value_type value_t;
    typedef Comparator<value_t> Comp;
    InPlaceSorter<Itr,Comp> sorter(first,last);
    sorter.sort();
}

template <class Itr, class Comp>
void sortInPlaceParallel( Itr const& first, Itr const& last, Comp const& comp )
{
    InPlaceParallelSorter<Itr,Comp> sorter(first,last,comp);
    sorter.sort();
}

template <class Itr>
void sortInPlaceParallel( Itr const& first, Itr const& last )
{
    typedef typename std::iterator_traits<Itr>::value_type value_t;
    typedef Comparator<value_t> Comp;
    InPlaceParallelSorter<Itr,Comp> sorter(first,last);
    sorter.sort();
}

#endif /* SORTINPLACE_H_ */
